/*---------------------------------------------------
 * file:    adldaMPICVB0.c
 * purpose: run simulated parallel topic model in MPI using CVB0 algorithm (see UAI 2009 paper)
 * inputs:  T (number of topics), iter (number of iterations), P (number of procs), seed, docword file
 * outputs: wp text file
 * version: 1.1
 * author:  asuncion@uci.edu
 * date:    11/12/2010
 *-------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <mpi.h>

void write_sparse(int nr, int nc, float **x, char *fname);
void sample_chain0 (int ntot, int W, int T, double alpha, double beta, int *w, int *d, int *c, float **z, float **wp, float **dp, float *ztot);
void printStats (int ntot, int W, int T, double alpha, double beta, int *w, int *d, int *z, int **wp, int **dp, int *ztot);
void chksum (int n, int T, int **x, int *sumx);
int **imat(int nr, int nc);
float **fmat(int nr, int nc);
int *ivec(int n);
double *dvec(int n);
float *fvec(int n);
int countntot(char *fname);
void randomassignment(int ntot, int T, int *w, int *d, int *c, float **z, float **wp, float **dp, float **dpSplitter, float *ztot);
void read_dw(char *fname, int *d, int *w, int *c, int *D, int *W);
void write_wp(char *fname, int **wp, int W, int T);

/*------------------------------------------
 * global variables
 *------------------------------------------ */
static int T; // number of topics
static int W; // number of unique words
static int D; // number of docs
static int ntot; // number of words in corpus

/*==========================================
 * main
 *========================================== */
int main(int argc, char* argv[])
{
  int rank,size,length,Pm1;
  char name[BUFSIZ];
  int i=0, j=0, k=0, h=0, proc, iter=0, seed, dump;
  int *w, **wSplit, *d, *c,**dSplit, **cSplit;
  int Dp, nstart, nend, dstart, dend;
  float **z,***zSplit,**wp,**sumwp,**wp0,**dwp,**dp,**dpSplitter,***dpSplit,*ztot,*ztot0;
  int *localntot; // keeps track of the local proc's number of words
  double alpha=0.1, beta=0.01;
  int P,t; // # of processors
  double nendTemp;
    

  char *fname;
  char outputFile[32];

  if (argc == 1) {
    fprintf(stderr, "usage: %s T iter P seed\n", argv[0]);
    exit(-1);
  }

  T    = atoi(argv[1]);
  iter = atoi(argv[2]);
  P = atoi(argv[3]);
  seed = atoi(argv[4]);
  fname = (char*)argv[5];

  assert(T>0);
  assert(iter>0);
  assert(alpha>0);
  assert(beta>0);
  assert(seed>0);

  ntot = countntot(fname);
  w = ivec(ntot);
  d = ivec(ntot);
  c = ivec(ntot);
  z = fmat(ntot,T);
  read_dw(fname, d, w, c, &D, &W);
  wp = fmat(W,T);
  sumwp = fmat(W,T);
  dp = fmat(D,T);
  dpSplitter = fmat(D,T);
  srand48(seed);  

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  MPI_Get_processor_name(name, &length);
  //printf("%s: hello world from process %d of %d\n", name, rank, size);

  ztot = fvec(T);
  randomassignment(ntot,T,w,d,c,z,wp,dp,dpSplitter,ztot);
  //chksum(W,T,wp,ztot);
  //chksum(D,T,dp,ztot);

  // Split over P procs
 
  Dp =(int) floor(D/P);

  nstart = 0;

  wSplit = (int**)calloc(P,sizeof(int*));
  dSplit = (int**)calloc(P,sizeof(int*));
  cSplit = (int**)calloc(P,sizeof(int*));
  zSplit = (float***)calloc(P,sizeof(float**));
  dpSplit = (float***)calloc(P,sizeof(float**));

  localntot = (int*)calloc(P,sizeof(int));


  nendTemp=0;
  for (proc = 0; proc < P; proc++)
  {
    dstart = proc*Dp;
    if (proc < P-1) {
      dend = (proc+1)*Dp - 1;
    }
    else
    {
      dend = D-1;
    }
    // Compute nend
    
    for (i = dstart; i <= dend; i++) {
      for (j = 0; j < T; j++)  {
        nendTemp += dpSplitter[i][j];
      }
    }
    // Subtract one since C's arrays are zero-based
    // actually I want to round here
    nend = (int) floor(nendTemp + 0.5) -1;

    // We are creating proc's local w, which will contain
    // the interval w(nstart:nend)
    wSplit[proc] = ivec(nend - nstart + 1);
    dSplit[proc] = ivec(nend - nstart + 1);
    cSplit[proc] = ivec(nend - nstart + 1);
    zSplit[proc] = fmat(nend - nstart + 1, T);
	
    for (i = nstart; i <= nend; i++) {
      wSplit[proc][i - nstart] = w[i];
      dSplit[proc][i - nstart] = d[i] - dstart;
      cSplit[proc][i - nstart] = c[i];

      for (t=0; t < T; t++)
      {
        zSplit[proc][i - nstart][t] = z[i][t];
      }

    }

    // Now we make proc's local dp matrix
    dpSplit[proc] = fmat(dend - dstart + 1, T);
    for (i = dstart; i <= dend; i++) {
      for (j = 0; j < T; j++) {
        dpSplit[proc][i-dstart][j] = dp[i][j];
      }
    }

    // set local ntot size of proc
    localntot[proc] = nend - nstart + 1;

    nstart = nend + 1;
  }



  wp0 = fmat(W,T);
  dwp = fmat(W,T);
  ztot0 = fvec(T);

  //if (wp0 == NULL || dwp == NULL || ztot0 == NULL) {
  //  printf("out of memory");
  //  abort();
  //}

  Pm1 = P - 1;

  // ***** Now we start to iterate 
  for (i = 1; i <= iter; i++) {

    if (rank == 0)
      printf("iter %d\n",i);

    // update wp0, dwp, and ztot0 all at once
    for (j = 0; j < T; j++) {
      ztot0[j] = 0;
      for (k = 0; k < W; k++) {
        // Copy wp to wp0
        wp0[k][j] = wp[k][j];
        // set dwp to be completely 0
        //dwp[k][j] = 0;
        ztot0[j] += wp0[k][j];
      }

      ztot[j] = ztot0[j];
    }

    //chksum(W,T,wp,ztot0);

    //memset(dwp[0], 0, W*T*sizeof(float));
    memset(sumwp[0], 0, W*T*sizeof(float));


    j = rank;
    // We are actually not doing sampling here but rather fast collapsed variational inference (CVB0),
    // which is a deterministic alternative to collapsed Gibbs Sampling
    sample_chain0(localntot[j],W,T,alpha,beta,wSplit[j],dSplit[j],cSplit[j],zSplit[j],wp,dpSplit[j],ztot);

    MPI_Allreduce(wp[0],sumwp[0],W*T,MPI_FLOAT,MPI_SUM,MPI_COMM_WORLD);

    for (j = 0; j < W; j++)
      for (k = 0; k < T; k++)
        wp[j][k] = sumwp[j][k] - Pm1*wp0[j][k];  // Update wp

  } // End of iterations
  
  // We need to write out wp

  if (rank==0) {
    sprintf(outputFile, "Nwt.txt");
    write_sparse(W, T, wp, outputFile);
  }

  sprintf(outputFile, "Ndt.%d.txt",rank);

  if (rank < P-1) {
    write_sparse(Dp, T, dpSplit[rank], outputFile);
  }
  else
  {
    write_sparse(D - (P-1)*Dp, T, dpSplit[rank], outputFile);
  }

  MPI_Finalize();

  return 0;
}

// This function just prints out w,d,z,wp,and dp for debugging purposes
void printStats(int localntot, int W, int T, double alpha, double beta, int *w, int *d, int *z, int **wp, int **dp, int *ztot)
{
  int i,j;
  printf("\nw: ");
  for (i =0; i < localntot; i++)
  {
    printf("%d ",w[i]);
  }
  printf("\nd: ");
  for (i =0; i < localntot; i++)
  {
    printf("%d ",d[i]);
  }
  printf("\nz: ");
  for (i =0; i < localntot; i++)
  {
    printf("%d ",z[i]);
  }
  printf("\nwp:\n");
  for (i = 0; i < W;i++)
  {
    for (j=0;j<T;j++)
    {
      printf("%d ",wp[i][j]);	
    }
    printf("\n");
  }	
  printf("\ndp:\n");
  for (i = 0; i < 2;i++)
  {
    for (j=0;j<T;j++)
    {
      printf("%d ",dp[i][j]);	
    }
    printf("\n");
  }
}

void sample_chain0 (int localntot, int W, int T, double alpha, double beta, int *w, int *d, int *c, float **z, float **wp, float **dp, float *ztot) //
{
  int i, t,wi,di, newlocal, cminus1;
  double totprob, maxprob, currprob, tmp;
  double *probs = dvec(T);
  double wbeta = W*beta;;



  for (i = 0; i < localntot; i++) {

    totprob = 0;
    wi = w[i];
    di = d[i];
    for (t=0; t < T; t++)  {
      ztot[t]=ztot[t] - z[i][t];
      wp[wi][t] = wp[wi][t] - z[i][t];
      dp[di][t] = dp[di][t] - z[i][t];

      probs[t] = (wp[wi][t] +  beta) *(dp[di][t]+  alpha)/(ztot[t]+  wbeta);
      totprob += probs[t];
    }

    cminus1 = c[i] - 1;

    for (t=0; t < T; t++)  {
      probs[t] = probs[t] / totprob;

      wp[wi][t] = wp[wi][t] - cminus1 * z[i][t] + c[i] * probs[t];
      dp[di][t] = dp[di][t] - cminus1 * z[i][t] + c[i] * probs[t];
      ztot[t] = ztot[t] - cminus1 * z[i][t] + c[i] * probs[t];

      z[i][t] = probs[t];

    }
  }




  free(probs);
}


void write_sparse(int nr, int nc, float **x, char *fname) //
{
  FILE *fp = fopen(fname,"w");
  int i, j;
  assert(fp);

  int total = 0;

  for (i = 0; i < nr; i++) {
    for (j = 0; j < nc; j++) {
      if (x[i][j] > 0) {
        total += 1; // # of distinct pairs
      }
    }
  }


  fprintf(fp, "%d\n", nr);
  fprintf(fp, "%d\n", nc);
  fprintf(fp, "%d\n", total);
  for (i = 0; i < nr; i++) 
    for (j = 0; j < nc; j++) 
      if (x[i][j] > 1e-6) fprintf(fp, "%d %d %.2f\n", i+1 , j+1 , x[i][j]);
  fclose(fp);
}





void chksum (int n, int T, int **x, int *sumx) //
{
  int i, t, sum;
  for (t = 0; t < T; t++) {
    sum = 0;
    for (i = 0; i < n; i++) sum += x[i][t];
    assert(sum==sumx[t]);
  }
}

int **imat(int nr, int nc) //
{
  int ntot = nr*nc;
  int *tmp = (int*) calloc(ntot,sizeof(int));
  int **x  = (int**)calloc(nr,sizeof(int*));
  int r;
  assert(tmp);
  assert(x);
  for (r = 0; r < nr; r++) x[r] = tmp + nc*r;
  return x;
}
  
int *ivec(int n) //
{
  int *x = (int*)calloc(n,sizeof(int));
  assert(x);
  return x;
}
  
double *dvec(int n) //
{
  double *x = (double*)calloc(n,sizeof(double));
  assert(x);
  return x;
}


float **fmat(int nr, int nc)
{
  int ntot = nr*nc;
  float *tmp = (float*) calloc(ntot,sizeof(float));
  float **x  = (float**)calloc(nr,sizeof(float*));
  int r;
  assert(tmp);
  assert(x);
  for (r = 0; r < nr; r++) x[r] = tmp + nc*r;
  return x;
}



float *fvec(int n)
{
  float *x = (float*)calloc(n,sizeof(float));
  assert(x);
  return x;
}



int countntot(char *fname) //
{
  int i, count, ntot = 0;
  char buf[BUFSIZ];
  FILE *fp = fopen(fname ,"r"); assert(fp);
  for (i = 0; i < 3; i++) fgets(buf, BUFSIZ, fp); // skip 3 header lines
  while (fscanf(fp, "%*d%*d%d", &count) != EOF) ntot += 1;
  fclose(fp);
  assert(ntot>0);
  return ntot;
}



void randomassignment(int ntot, int T, int *w, int *d, int *c, float **z, float **wp, float **dp, float **dpSplitter, float *ztot)
{
  int i, t,wi,di;
  double totprob;
  float tmp;
  float tmp2 = 1/(double)T;
  for (i = 0; i < ntot; i++) {
    wi=w[i];
    di=d[i];
   
    totprob = 0;
    for (t=0; t < T; t++) {
      z[i][t] = drand48();
      totprob = totprob + z[i][t];
    }

   
    for (t=0; t < T; t++) {
      z[i][t] = z[i][t] / totprob;
      //z[i][t] = tmp2;
      tmp = c[i] * z[i][t];
      wp[wi][t] =  wp[wi][t] + tmp;
      dp[di][t] = dp[di][t] + tmp;
      dpSplitter[di][t] = dpSplitter[di][t] +  z[i][t];
      ztot[t] = ztot[t] + tmp;
    }

    //c[i] = c[i]-1;
  }
}




void read_dw(char *fname, int *d, int *w, int *c, int *D, int *W) //
{
  int i,wt,dt,ct,count,nnz;
  FILE *fp = fopen(fname ,"r"); assert(fp);
  fscanf(fp,"%d", D);    assert(*D>0);
  fscanf(fp,"%d", W);    assert(*W>0);
  fscanf(fp,"%d", &nnz); assert(nnz>0);
  i=0;
  while (fscanf(fp, "%d%d%d", &dt, &wt, &ct) != EOF) {
      w[i] = wt-1;
      d[i] = dt-1;
      c[i] = ct;
      i=i+1;
  }
  fclose(fp);
}



void write_wp(char *fname, int **wp, int W, int T)
{
  int i, j, total;

  FILE *fp = fopen(fname,"w"); assert(fp);

  total = 0;

  for (i = 0; i < W; i++) {
    for (j = 0; j < T; j++) {
      if (wp[i][j] > 0) {
        total += 1; // # of distinct pairs
      }
    }
  }

  fprintf(fp,"%d\n",W);
  fprintf(fp,"%d\n",T);
  fprintf(fp,"%d\n",total);

  for (i = 0; i < W; i++) {
    for (j = 0; j < T; j++) {
      if (wp[i][j] > 0) {
        fprintf(fp,"%d %d %d\n",i+1,j+1,wp[i][j]);
      }
    }
  }
  fclose(fp);

}
